import { Connection, PublicKey } from '@solana/web3.js';
import { useEffect, useMemo, useState } from 'react';
import { MintLayout, AccountLayout } from '@solana/spl-token';
import { utils, models, contexts } from '@oyster/common';
import { PoolInfo } from '../models';
const { programIds } = utils;
const {
  TokenSwapLayout,
  TokenSwapLayoutLegacyV0: TokenSwapLayoutV0,
  TokenSwapLayoutV1,
} = models;
const { useConnection } = contexts.Connection;
const { cache, getMultipleAccounts, TokenAccountParser } = contexts.Accounts;

export const LIQUIDITY_PROVIDER_FEE = 0.003;
export const SERUM_FEE = 0.0005;

const getHoldings = (connection: Connection, accounts: string[]) => {
  return accounts.map(acc => cache.query(connection, new PublicKey(acc)));
};

const toPoolInfo = (item: any, program: PublicKey) => {
  return {
    pubkeys: {
      account: item.pubkey,
      program: program,
      mint: item.data.tokenPool,
      holdingMints: [] as PublicKey[],
      holdingAccounts: [item.data.tokenAccountA, item.data.tokenAccountB],
    },
    legacy: false,
    raw: item,
  } as PoolInfo;
};

export const usePools = () => {
  const connection = useConnection();
  const [pools, setPools] = useState<PoolInfo[]>([]);

  // initial query
  useEffect(() => {
    setPools([]);

    const queryPools = async (swapId: PublicKey, isLegacy = false) => {
      let poolsArray: PoolInfo[] = [];
      (await connection.getProgramAccounts(swapId))
        .filter(
          item =>
            item.account.data.length === TokenSwapLayout.span ||
            item.account.data.length === TokenSwapLayoutV1.span ||
            item.account.data.length === TokenSwapLayoutV0.span,
        )
        .map(item => {
          let result = {
            data: undefined as any,
            account: item.account,
            pubkey: item.pubkey,
            init: async () => {},
          };

          const layout =
            item.account.data.length === TokenSwapLayout.span
              ? TokenSwapLayout
              : item.account.data.length === TokenSwapLayoutV1.span
              ? TokenSwapLayoutV1
              : TokenSwapLayoutV0;

          // handling of legacy layout can be removed soon...
          if (layout === TokenSwapLayoutV0) {
            result.data = layout.decode(item.account.data);
            let pool = toPoolInfo(result, swapId);
            pool.legacy = isLegacy;
            poolsArray.push(pool as PoolInfo);

            result.init = async () => {
              try {
                // TODO: this is not great
                // Ideally SwapLayout stores hash of all the mints to make finding of pool for a pair easier
                const holdings = await Promise.all(
                  getHoldings(connection, [
                    result.data.tokenAccountA,
                    result.data.tokenAccountB,
                  ]),
                );

                pool.pubkeys.holdingMints = [
                  holdings[0].info.mint,
                  holdings[1].info.mint,
                ] as PublicKey[];
              } catch (err) {
                console.log(err);
              }
            };
          } else {
            result.data = layout.decode(item.account.data);

            let pool = toPoolInfo(result, swapId);
            pool.legacy = isLegacy;
            pool.pubkeys.feeAccount = result.data.feeAccount;
            pool.pubkeys.holdingMints = [
              result.data.mintA,
              result.data.mintB,
            ] as PublicKey[];

            poolsArray.push(pool as PoolInfo);
          }

          return result;
        });

      const toQuery = poolsArray
        .map(
          p =>
            [
              ...p.pubkeys.holdingAccounts.map(h => h.toBase58()),
              ...p.pubkeys.holdingMints.map(h => h.toBase58()),
              p.pubkeys.feeAccount?.toBase58(), // used to calculate volume aproximation
              p.pubkeys.mint.toBase58(),
            ].filter(p => p) as string[],
        )
        .flat();

      // This will pre-cache all accounts used by pools
      // All those accounts are updated whenever there is a change
      await getMultipleAccounts(connection, toQuery, 'single').then(
        ({ keys, array }) => {
          return array.map((obj, index) => {
            const pubKey = keys[index];
            if (obj.data.length === AccountLayout.span) {
              return cache.add(pubKey, obj, TokenAccountParser);
            } else if (obj.data.length === MintLayout.span) {
              if (!cache.getMint(pubKey)) {
                return cache.addMint(new PublicKey(pubKey), obj);
              }
            }

            return obj;
          }) as any[];
        },
      );

      return poolsArray;
    };
    Promise.all([
      queryPools(programIds().swap),
      ...programIds().swap_legacy.map(leg => queryPools(leg, true)),
    ]).then(all => {
      setPools(all.flat());
    });
  }, [connection]);

  useEffect(() => {
    const subID = connection.onProgramAccountChange(
      programIds().swap,
      async info => {
        const pubkey =
          typeof info.accountId === 'string'
            ? new PublicKey((info.accountId as unknown) as string)
            : info.accountId;
        const id = pubkey.toBase58();

        if (info.accountInfo.data.length === programIds().swapLayout.span) {
          const account = info.accountInfo;
          const updated = {
            data: programIds().swapLayout.decode(account.data),
            account: account,
            pubkey,
          };

          const index =
            pools && pools.findIndex(p => p.pubkeys.account.toBase58() === id);
          if (index && index >= 0 && pools) {
            // TODO: check if account is empty?

            const filtered = pools.filter((p, i) => i !== index);
            setPools([...filtered, toPoolInfo(updated, programIds().swap)]);
          } else {
            let pool = toPoolInfo(updated, programIds().swap);

            pool.pubkeys.feeAccount = updated.data.feeAccount;
            pool.pubkeys.holdingMints = [
              updated.data.mintA,
              updated.data.mintB,
            ] as PublicKey[];

            setPools([...pools, pool]);
          }
        }
      },
      'singleGossip',
    );

    return () => {
      connection.removeProgramAccountChangeListener(subID);
    };
  }, [connection, pools]);

  return { pools };
};

export const usePoolForBasket = (mints: (string | undefined)[]) => {
  const connection = useConnection();
  const { pools } = usePools();
  const [pool, setPool] = useState<PoolInfo>();
  const sortedMints = useMemo(() => [...mints].sort(), [...mints]); // eslint-disable-line
  useEffect(() => {
    (async () => {
      // reset pool during query
      setPool(undefined);
      let matchingPool = pools
        .filter(p => !p.legacy)
        .filter(p =>
          p.pubkeys.holdingMints
            .map(a => a.toBase58())
            .sort()
            .every((address, i) => address === sortedMints[i]),
        );

      for (let i = 0; i < matchingPool.length; i++) {
        const p = matchingPool[i];

        const account = await cache.query(
          connection,
          p.pubkeys.holdingAccounts[0],
        );

        if (!account.info.amount.eqn(0)) {
          setPool(p);
          return;
        }
      }
    })();
  }, [connection, sortedMints, pools]);

  return pool;
};

function estimateProceedsFromInput(
  inputQuantityInPool: number,
  proceedsQuantityInPool: number,
  inputAmount: number,
): number {
  return (
    (proceedsQuantityInPool * inputAmount) / (inputQuantityInPool + inputAmount)
  );
}

function estimateInputFromProceeds(
  inputQuantityInPool: number,
  proceedsQuantityInPool: number,
  proceedsAmount: number,
): number | string {
  if (proceedsAmount >= proceedsQuantityInPool) {
    return 'Not possible';
  }

  return (
    (inputQuantityInPool * proceedsAmount) /
    (proceedsQuantityInPool - proceedsAmount)
  );
}

export enum PoolOperation {
  Add,
  SwapGivenInput,
  SwapGivenProceeds,
}

export async function calculateDependentAmount(
  connection: Connection,
  independent: string,
  amount: number,
  pool: PoolInfo,
  op: PoolOperation,
): Promise<number | string | undefined> {
  const poolMint = await cache.queryMint(connection, pool.pubkeys.mint);
  const accountA = await cache.query(
    connection,
    pool.pubkeys.holdingAccounts[0],
  );
  const amountA = accountA.info.amount.toNumber();

  const accountB = await cache.query(
    connection,
    pool.pubkeys.holdingAccounts[1],
  );
  let amountB = accountB.info.amount.toNumber();

  if (!poolMint.mintAuthority) {
    throw new Error('Mint doesnt have authority');
  }

  if (poolMint.supply.eqn(0)) {
    return;
  }

  let offsetAmount = 0;
  const offsetCurve = pool.raw?.data?.curve?.offset;
  if (offsetCurve) {
    offsetAmount = offsetCurve.token_b_offset;
    amountB = amountB + offsetAmount;
  }

  const mintA = await cache.queryMint(connection, accountA.info.mint);
  const mintB = await cache.queryMint(connection, accountB.info.mint);

  if (!mintA || !mintB) {
    return;
  }

  const isFirstIndependent = accountA.info.mint.toBase58() === independent;
  const depPrecision = Math.pow(
    10,
    isFirstIndependent ? mintB.decimals : mintA.decimals,
  );
  const indPrecision = Math.pow(
    10,
    isFirstIndependent ? mintA.decimals : mintB.decimals,
  );
  const indAdjustedAmount = amount * indPrecision;

  let indBasketQuantity = isFirstIndependent ? amountA : amountB;

  let depBasketQuantity = isFirstIndependent ? amountB : amountA;

  var depAdjustedAmount;

  const constantPrice = pool.raw?.data?.curve?.constantPrice;
  if (constantPrice) {
    depAdjustedAmount = (amount * depPrecision) / constantPrice.token_b_price;
  } else {
    switch (+op) {
      case PoolOperation.Add:
        depAdjustedAmount =
          (depBasketQuantity / indBasketQuantity) * indAdjustedAmount;
        break;
      case PoolOperation.SwapGivenProceeds:
        depAdjustedAmount = estimateInputFromProceeds(
          depBasketQuantity,
          indBasketQuantity,
          indAdjustedAmount,
        );
        break;
      case PoolOperation.SwapGivenInput:
        depAdjustedAmount = estimateProceedsFromInput(
          indBasketQuantity,
          depBasketQuantity,
          indAdjustedAmount,
        );
        break;
    }
  }

  if (typeof depAdjustedAmount === 'string') {
    return depAdjustedAmount;
  }
  if (depAdjustedAmount === undefined) {
    return undefined;
  }
  return depAdjustedAmount / depPrecision;
}
